//-----------------------------------------------------------------------------
// hill climbing algorithm for QAP 
//-----------------------------------------------------------------------------

// standard includes
#include <stdexcept>  // runtime_error 
#include <iostream>   // cout
#include <strstream>  // ostrstream, istrstream
#include <fstream>
#include <string.h>

// the general include for eo
#include "EO.h"
#include "ga.h"
#include "es.h"

using namespace std;

// includes from Moving Object
//-----------------------------------------------------------------------------
#include <eoHC.h>
#include <eoFirstImprSelect.h>
#include <eoBestImprSelect.h>
#include <eoRandImprSelect.h>

// REPRESENTATION
//-----------------------------------------------------------------------------
#include "perm.h"
#include "initPerm.h"

// define your individuals
typedef Perm<double> Indi;   


// specific includes
//-----------------------------------------------------------------------------
#include "funcQAP.h"   // fitness function
#include "incrEvalQAP.h"   // incremental evaluation

#include "permMove.h"       // movement for permutation
#include "initPermMove.h"   // initial movement for permutation
#include "nextPermMove.h"   // next movement for permutation


// GENERAL
//-----------------------------------------------------------------------------

void main_function(int argc, char **argv)
{
  // First define a parser from the command-line arguments
  eoParser parser(argc, argv);

  // For each parameter, define Parameter, read it through the parser,
  // and assign the value to the variable

  eoValueParam<uint32> seedParam(time(0), "seed", "Random number seed", 'S');
  parser.processParam( seedParam );
  unsigned seed = seedParam.value();
  
  // description of genotype
  eoValueParam<unsigned int> vecSizeParam(8, "vecSize", "Genotype size",'V');
  parser.processParam( vecSizeParam, "Representation" );
  unsigned vecSize = vecSizeParam.value();


  char fileQAPname[256]; // default value
  strcpy(fileQAPname, "nug30.dat");
  eoValueParam<char *> fileParam(fileQAPname, "file", "file name of QAP instance", 'F');
  parser.processParam( fileParam, "Representation" );
  
  // init and stop
  eoValueParam<string> loadNameParam("", "Load","A save file to restart from",'L');
  parser.processParam( loadNameParam, "Persistence" );
  string loadName = loadNameParam.value();
  
  eoValueParam<unsigned int> maxGenParam(100, "maxGen", "Maximum number of generations",'G');
  parser.processParam( maxGenParam, "Stopping criterion" );
  unsigned maxGen = maxGenParam.value();
  
  char fileStat[512];
  strcpy(fileStat, "stats.dat");
  eoValueParam<char *> fileStatParam(fileStat,"stat","A file to export fitness");
  parser.processParam( fileStatParam, "Persistence" );

  // the name of the "status" file where all actual parameter values will be saved
  string str_status = parser.ProgramName() + ".status"; // default value
  eoValueParam<string> statusParam(str_status.c_str(), "status","Status file",'S');
  parser.processParam( statusParam, "Persistence" );

  // do the following AFTER ALL PARAMETERS HAVE BEEN PROCESSED
  // i.e. in case you need parameters somewhere else, postpone these
  if (parser.userNeedsHelp())
    {
      parser.printHelp(cout);
       exit(1);
    }
  if (statusParam.value() != "")
    {
      ofstream os(statusParam.value().c_str());
      os << parser;		// and you can use that file as parameter file
    }


// GENERAL
  //////////////////////////
  //  Random seed
  //////////////////////////
  //reproducible random seed: if you don't change SEED above, 
  // you'll aways get the same result, NOT a random run
  rng.reseed(seed);

  // EVAL
  // Fitness function
  
  FuncQAP<Indi> eval(fileQAPname);

// INIT
  ////////////////////////////////
  // Initilisation of population
  ////////////////////////////////
  // Either load or initialize
  // create an empty pop
  Indi solution;

  // create a state for reading
  eoState inState;		
  // a state for loading - WITHOUT the parser
  // register the rng and the pop in the state, so they can be loaded,
  // and the present run will be the exact conitnuation of the saved run
  // eventually with different parameters
  inState.registerObject(rng);
    
  if (loadName != "")
    {
      inState.load(loadName); //  load the pop and the rng
      // the fitness is read in the file: 
      // do only evaluate the pop if the fitness has changed
    }
  else
    {
      rng.reseed(seed);
      // a Indi random initializer
      PermInit<Indi> random(vecSize);
	
      // Init 
      random(solution);      

      // and evaluate 
      eval(solution);
    } 

  // OUTPUT    
  cout << "Initial Solution" << endl << solution << endl;

  // incremental evaluation
  IncrEvalQAP< PermMove<Indi> > incrEval(&eval); 

  // init movement : first solution of the neighborhood
  InitPermMove<Indi> initMove;

  // Next movement to explore the neighborhood
  NextPermMove<Indi> nextMove;

  // selection of move

  //eoFirstImprSelect<PermMove> selectMove;
  eoBestImprSelect< PermMove<Indi> > selectMove ;
  //eoRandImprSelect<PermMove> selectMove;  


// GENERATION
  /////////////////////////////////////////
  // the algorithm
  ////////////////////////////////////////

  eoHC< PermMove<Indi> > hill_climb(initMove, nextMove, incrEval, selectMove, eval) ;

  // Apply algo to solution - that's it!
  hill_climb(solution) ;
  
// OUTPUT
  cout << "FINAL solution" << endl << solution << endl;

  fstream file(fileStat, ios::app);
  if (file.is_open()) {
    file << solution <<endl;
    file.close();
  } else
    cerr << "impossible to open " << fileStat <<endl;

// GENERAL
}

// A main that catches the exceptions

int main(int argc, char **argv)
{

    try
    {
        main_function(argc, argv);
    }
    catch(exception& e)
    {
        cout << "Exception: " << e.what() << '\n';
    }

    return 1;
}
