//-----------------------------------------------------------------------------
// agNK.cpp
//-----------------------------------------------------------------------------
// standard includes
#include <stdexcept>  // runtime_error 
#include <iostream>   // cout
#include <strstream>  // ostrstream, istrstream
#include <fstream>
#include <string.h>

// the general include for eo
#include "EO.h"
#include "ga.h"
#include "es.h"

using namespace std;

#include "perm.h"
#include "initPerm.h"

// REPRESENTATION
//-----------------------------------------------------------------------------
// define your individuals
typedef Perm<double> Indi;   

// includes
//-----------------------------------------------------------------------------
#include "funcQAP.h"
#include "xoverQAP.h"
#include "mutationQAP.h"

// GENERAL
//-----------------------------------------------------------------------------

void main_function(int argc, char **argv)
{
  // First define a parser from the command-line arguments
  eoParser parser(argc, argv);

  // For each parameter, define Parameter, read it through the parser,
  // and assign the value to the variable

  eoValueParam<uint32_t> seedParam(time(0), "seed", "Random number seed", 'S');
  parser.processParam( seedParam );
  unsigned seed = seedParam.value();
  
  // description of genotype
  eoValueParam<unsigned int> vecSizeParam(8, "vecSize", "Genotype size",'V');
  parser.processParam( vecSizeParam, "Representation" );
  unsigned vecSize = vecSizeParam.value();


  char fileQAPname[256]; // default value
  strcpy(fileQAPname, "nug30.dat");
  eoValueParam<char *> fileParam(fileQAPname, "file", "file name of QAP instance", 'F');
  parser.processParam( fileParam, "Representation" );
  
  // parameters for evolution engine
  eoValueParam<unsigned int> popSizeParam(10, "popSize", "Population size",'P');
  parser.processParam( popSizeParam, "Evolution engine" );
  unsigned popSize = popSizeParam.value();
  
  eoValueParam<unsigned int> tSizeParam(2, "tSize", "Tournament size",'T');
  parser.processParam( tSizeParam, "Evolution Engine" );
  unsigned tSize = tSizeParam.value();

  // init and stop
  eoValueParam<string> loadNameParam("", "Load","A save file to restart from",'L');
  parser.processParam( loadNameParam, "Persistence" );
  string loadName = loadNameParam.value();
  
  eoValueParam<unsigned int> maxGenParam(100, "maxGen", "Maximum number of generations",'G');
  parser.processParam( maxGenParam, "Stopping criterion" );
  unsigned maxGen = maxGenParam.value();
  
  eoValueParam<unsigned int> minGenParam(100, "minGen", "Minimum number of generations",'g');
  parser.processParam( minGenParam, "Stopping criterion" );
  unsigned minGen = minGenParam.value();
  
  eoValueParam<unsigned int> steadyGenParam(100, "steadyGen", "Number of generations with no improvement",'s');
  parser.processParam( steadyGenParam, "Stopping criterion" );
  unsigned steadyGen = steadyGenParam.value();
  
  // operators probabilities at the algorithm level
  eoValueParam<double> pCrossParam(0.6, "pCross", "Probability of Crossover", 'C'); 
  parser.processParam( pCrossParam, "Genetic Operators" );
  double pCross = pCrossParam.value();
  
  eoValueParam<double> pMutParam(0.1, "pMut", "Probability of Mutation", 'M');
  parser.processParam( pMutParam, "Genetic Operators" );
  double pMut = pMutParam.value();
  
  char fileStat[512];
  strcpy(fileStat, "stats.dat");
  eoValueParam<char *> fileStatParam(fileStat,"stat","A file to export fitness");
  parser.processParam( fileStatParam, "Persistence" );

  // the name of the "status" file where all actual parameter values will be saved
  string str_status = parser.ProgramName() + ".status"; // default value
  eoValueParam<string> statusParam(str_status.c_str(), "status","Status file",'S');
  parser.processParam( statusParam, "Persistence" );

  // do the following AFTER ALL PARAMETERS HAVE BEEN PROCESSED
  // i.e. in case you need parameters somewhere else, postpone these
  if (parser.userNeedsHelp())
    {
      parser.printHelp(cout);
      exit(1);
    }
  if (statusParam.value() != "")
    {
      ofstream os(statusParam.value().c_str());
      os << parser;		// and you can use that file as parameter file
    }


// GENERAL
  //////////////////////////
  //  Random seed
  //////////////////////////
  //reproducible random seed: if you don't change SEED above, 
  // you'll aways get the same result, NOT a random run
  rng.reseed(seed);

  // EVAL
  // Fitness function
  
  FuncQAP<Indi> eval(fileQAPname);

// INIT
  ////////////////////////////////
  // Initilisation of population
  ////////////////////////////////
  // Either load or initialize
  // create an empty pop
  eoPop<Indi> pop;
  // create a state for reading
  eoState inState;		// a state for loading - WITHOUT the parser
  // register the rng and the pop in the state, so they can be loaded,
  // and the present run will be the exact conitnuation of the saved run
  // eventually with different parameters
  inState.registerObject(rng);
  inState.registerObject(pop);
    
  if (loadName != "")
    {
      inState.load(loadName); //  load the pop and the rng
      // the fitness is read in the file: 
      // do only evaluate the pop if the fitness has changed
    }
  else
    {
      rng.reseed(seed);
      // a Indi random initializer
      // based on boolean_generator class (see utils/rnd_generator.h)
      PermInit<Indi> random(vecSize);
	
      // Init pop from the randomizer: need to use the append function
      pop.append(popSize, random);      
      // and evaluate pop (STL syntax)    
      apply<Indi>(eval, pop);
    } // end of initializatio of the population

// OUTPUT    
  // sort pop for pretty printout
  //  pop.sort();
  // Print (sorted) intial population (raw printout)
  //  cout << "Initial Population" << endl << pop << endl;

// ENGINE
  /////////////////////////////////////
  // selection and replacement
  ////////////////////////////////////
// SELECT
  // The robust tournament selection
  eoDetTournamentSelect<Indi> selectOne(tSize);       // tSize in [2,POPSIZE]
  // is now encapsulated in a eoSelectPerc (stands for Percentage)
  eoSelectPerc<Indi> select(selectOne); 

// REPLACE
  // And we now have the full slection/replacement - though with 
  // the same generational replacement at the moment :-)
  eoGenerationalReplacement<Indi> replaceGen; 

  eoWeakElitistReplacement<Indi> replace(replaceGen);

// OPERATORS
  //////////////////////////////////////
  // The variation operators
  //////////////////////////////////////
// CROSSOVER
  // 1-point crossover for bitstring
  XoverQAP<Indi> xover;

// MUTATION

  // standard bit-flip mutation for bitstring
  MutationQAP<Indi> mutation;

  // The operators are  encapsulated into an eoTRansform object
  eoSGATransform<Indi> transform(xover, pCross, mutation, pMut);

// STOP
  //////////////////////////////////////
  // termination condition see FirstBitEA.cpp
  /////////////////////////////////////
  eoGenContinue<Indi> continuator(maxGen);
  /*
  eoSteadyFitContinue<Indi> steadyCont(minGen, steadyGen);
  eoFitContinue<Indi> fitCont(vecSize);
  eoCombinedContinue<Indi> continuator(genCont);
  continuator.add(steadyCont);
  continuator.add(fitCont);
  */
  
// CHECKPOINT
  // but now you want to make many different things every generation 
  // (e.g. statistics, plots, ...).
  // the class eoCheckPoint is dedicated to just that:

  // Declare a checkpoint (from a continuator: an eoCheckPoint 
  // IS AN eoContinue and will be called in the loop of all algorithms)
  eoCheckPoint<Indi> checkpoint(continuator);

  // Create a counter parameter
  eoValueParam<unsigned> generationCounter(0, "Gen.");

  // Create an incrementor (sub-class of eoUpdater). Note that the 
  // parameter's value is passed by reference, 
  // so every time the incrementer is updated (every generation),
  // the data in generationCounter will change.
  eoIncrementor<unsigned> increment(generationCounter.value());

  // Add it to the checkpoint, 
  // so the counter is updated (here, incremented) every generation
  checkpoint.add(increment);

  // now some statistics on the population:
  // Best fitness in population
  eoBestFitnessStat<Indi> bestStat;
  // Second moment stats: average and stdev
  eoSecondMomentStats<Indi> SecondStat;

  // Add them to the checkpoint to get them called at the appropriate time
  checkpoint.add(bestStat);
  checkpoint.add(SecondStat);

  // A file monitor: will print parameters to ... a File, yes, you got it!
  eoFileMonitor fileMonitor(fileStat, " ");
  
  // the checkpoint mechanism can handle multiple monitors
  checkpoint.add(fileMonitor);

  // the fileMonitor can monitor parameters, too, but you must tell it!
  fileMonitor.add(generationCounter);
  fileMonitor.add(bestStat);
  fileMonitor.add(SecondStat);

  // Last type of item the eoCheckpoint can handle: state savers:
  eoState outState;
  // Register the algorithm into the state (so it has something to save!!)
  outState.registerObject(parser);
  outState.registerObject(pop);
  outState.registerObject(rng);

  /*
  // and feed the state to state savers
  // save state every 100th  generation
  eoCountedStateSaver stateSaver1(20, outState, "generation"); 
  // save state every 1 seconds 
  eoTimedStateSaver   stateSaver2(1, outState, "time"); 
  
  // Don't forget to add the two savers to the checkpoint
  checkpoint.add(stateSaver1);
  checkpoint.add(stateSaver2);
  // and that's it for the (control and) output
  */

// GENERATION
  /////////////////////////////////////////
  // the algorithm
  ////////////////////////////////////////

  // Easy EA requires
  // stopping criterion, eval, selection, transformation, replacement
  eoEasyEA<Indi> gga(checkpoint, eval, select, transform, replace);

  // Apply algo to pop - that's it!
  gga(pop);
  
// OUTPUT
  // Print (sorted) intial population
  //  pop.sort();
  //cout << "FINAL Population\n" << pop << endl;
// GENERAL
}

// A main that catches the exceptions

int main(int argc, char **argv)
{

    try
    {
        main_function(argc, argv);
    }
    catch(exception& e)
    {
        cout << "Exception: " << e.what() << '\n';
    }

    return 1;
}
